/* asaai.cpp
 * Copyright (C) 2004 Djukic Marko, Pernus Klemen, Zagar Miha
 * For conditions of distribution and use, see the copyright notice in main.cpp
 */
#include "asaai.h"

#include <iostream>
using namespace std;

//konstruktor
Asaai::Asaai(FXCanvas *_canvas, FXButton *_text) {
	canvas = _canvas;
	text_button = _text;

	ai = new AI();
	//prepoznavalec je odvisen od ai	
	recognizer = new Recognizer(ai);

	//delo z omejitvam odvisen od ai , in obeh seznamov
	constraint_manager = new Constraint_Manager(ai,&object_list,&constraint_list);
	
	//miskino stanje
	mouse_state = new Mouse();
	//omejitve so vidne
	show_constraints = 1;
}

AI *Asaai::get_ai() {
	return ai;
}

Recognizer *Asaai::get_recognizer() {
	return recognizer;
}


void Asaai::clear() {
	list <Object*>::iterator iter_obj;
	Object *obj;

	object_list.clear();
	constraint_list.clear();

	text_button->disable();
}

void Asaai::unselect_constraints() {
	list <Constraint*>::iterator iter_cn;

	for (iter_cn = constraint_list.begin(); iter_cn != constraint_list.end(); iter_cn++)
	{
		(*iter_cn)->set_selected(0);
	}
}

void Asaai::unselect_objects() {
	list <Object*>::iterator iter_obj;

	for (iter_obj = object_list.begin(); iter_obj != object_list.end(); iter_obj++)
	{
		(*iter_obj)->set_selected(0);
	}

	text_button->disable();
}

void Asaai::on_lbutton_down(Point *pozicija) {
	//zadeva je OK samo ce prej ni bilo nic pritisnjenega
	if (mouse_state->pressed == 0) { //je OK
		//odoznacim objekte
		unselect_objects();
		//odoznacim omejitve
		unselect_constraints();
		//ustavrim tracerja
		last_input = new Trace();
		//popravim koordinate miske
		mouse_state->pressed = 1;
		mouse_state->button = 1;
		mouse_state->double_click = 0;
		mouse_state->start->move_to(pozicija);
		mouse_state->previous->move_to(pozicija);
		mouse_state->current->move_to(pozicija);
	
		//narisem spremembo na ekran
		on_paint();
		update_screen();
	}
}

void Asaai::on_lbutton_up(Point *pozicija) {
	Object *recognized_object;

	//zadeva je OK samo ce je bil enojni levi klik
	if (mouse_state->pressed && mouse_state->button == 1 && mouse_state->double_click == 0) {

		//popravim koordinate miska
		mouse_state->pressed = 0;
		mouse_state->button = 0;
		mouse_state->double_click = 0;
		mouse_state->previous->move_to(mouse_state->current);
		mouse_state->current->move_to(pozicija);

		//narisem spremembo na ekran
		update_screen();

		if (recognizer->recognize(last_input) == 1) { //prepoznavaclec je nasel vsaj en objekt
			while ((recognized_object = recognizer->get_object()) != NULL) { //dokler obstajajo objekti
				if (((Line*)recognized_object)->get_length() > 0) {
					recognized_object->set_object_color(mouse_state->color);
					recognized_object->set_object_width(mouse_state->width);
					constraint_manager->add_object(recognized_object); //objekt posredujem managerju za pravilno umestitev
				}
			}
			constraint_manager->add_object(NULL);
		}

		//si zbrisem trenutni input
		delete last_input;

		//izrisem kar je trenutno narsano
		on_paint();
	}
}

void Asaai::on_rbutton_down(Point *pozicija) {
	list <Object*>::iterator iter_obj;
	list <Constraint*>::iterator iter_cn;

	//zadeva je OK samo ce ni bilo nic pritisnjenaga
	if (mouse_state->pressed == 0) {

		//popravim koordinate miska
		mouse_state->pressed = 1;
		mouse_state->button = 2;
		mouse_state->double_click = 0;
		mouse_state->start->move_to(pozicija);
		mouse_state->previous->move_to(pozicija);
		mouse_state->current->move_to(pozicija);
	
		cout << "desni klik" << endl;
		for (iter_obj = object_list.begin(); iter_obj != object_list.end(); iter_obj++) //dokler so objekti
		{
			//cout << "smo v blizini " << (*iter_obj)->get_age() << endl;
			if ((*iter_obj)->on_object(pozicija)) //ce je klik na nek objekt
			{
				//najprej odoznacim objekte ... ce je kaksen ostal oznacen ... to je potrebno
				//narediti tu, da pri kliku na omejitev vem ce je kak ojekt izbran
				unselect_objects();
				(*iter_obj)->set_selected(1);//oznacim
				text_button->enable();
				(*iter_obj)->on_vertex(pozicija);//ce sem na tockau .. postavi closest vertex
				mouse_state->holding_object_vertex = (*iter_obj)->get_closest_vertex();//dobim kje drzim zadevo
				mouse_state->holding_object = (*iter_obj);//in primem
				on_paint();//narisem spremembo
				return;
			}
		}

		if (show_constraints == 1) { //omejitve so vidne, torej jih moram gledat ce je izbrana
			for (iter_cn = constraint_list.begin(); iter_cn != constraint_list.end(); iter_cn++)
			{
				if ((*iter_cn)->on_object(pozicija))
				{
					(*iter_cn)->set_selected(1);
					mouse_state->holding_constraint = (*iter_cn);
					//narisem spremembo
					on_paint();
					return; //in pobegnem iz funkcije
				}
			}
		}
	}
}

void Asaai::on_rbutton_up(Point *pozicija) {

	//zadeva je OK samo ce je bil prej pritisnjen desni gumb
	if (mouse_state->pressed && mouse_state->button == 2 && mouse_state->double_click == 0) {
	
		//popravim koordinate miska
		mouse_state->pressed = 0;
		mouse_state->button = 0;
		mouse_state->double_click = 0;
		mouse_state->previous->move_to(mouse_state->current);
		mouse_state->current->move_to(pozicija);

		if (show_constraints == 1) { //omejitve so vidne ... je sansa da je drzal omejiev
			if (mouse_state->holding_constraint != NULL) { //je drzal omejitev
				if (mouse_state->holding_constraint->is_selected()) { //ce je se vedno izbrana (jo drzim)
					//jo posredujem managerju za izbris
					constraint_manager->delete_constraint(mouse_state->holding_constraint);

				}
			}
		}
		//spustim vse omejitve
		mouse_state->holding_constraint = NULL;
		mouse_state->holding_object = NULL;

		
		if (constraint_manager->obj_changed != NULL)
		{
			constraint_manager->clear_visited_on_constraints();
			constraint_manager->recheck_objects(constraint_manager->obj_changed);
			constraint_manager->split_changed_objects();
			constraint_manager->recheck_constraints();
			constraint_manager->obj_changed = NULL;
		}


		//odoznacim omejitve
		unselect_constraints();
		//narisem spremembe
		on_paint();
	}
}

void Asaai::on_lbutton_double(Point *pozicija) {
	//oznacim si dvoklik
	mouse_state->double_click = 1;
}

void Asaai::on_rbutton_double(Point *pozicija) {
	list <Object*>::iterator iter_obj;
	
	mouse_state->double_click = 1;

	for (iter_obj = object_list.begin(); iter_obj != object_list.end(); iter_obj++)
	{
		if ((*iter_obj)->on_object(pozicija)) { //ce je dvoklik na nek objekt
			constraint_manager->delete_object((*iter_obj));
			//pobegnem iz zanke
			break;
		}
	}
	
	//narisem spremembo
	on_paint();
}

void Asaai::on_mouse_move(Point *pozicija) {
	Point *dp;
	
	if (mouse_state->pressed) { //samo ce je pritisnjen kateri od gumbov
		mouse_state->previous->move_to(mouse_state->current);
		mouse_state->current->move_to(pozicija);

		switch (mouse_state->button) { //glede na pritisnjen gumb
			case 1: //levi gumb .. se rise po ekranu
				update_screen();
				break;
			case 2: //desni gumb .. se premika objekt ali brise omejitev
				if (mouse_state->holding_object != NULL) { //pomeni da drzim objekt v roki (desni klik)
					dp = new Point(mouse_state->current->x - mouse_state->previous->x, mouse_state->current->y - mouse_state->previous->y);
					switch (mouse_state->holding_object_vertex) {
						case 0:  //drzim na objeku
							constraint_manager->move_object_by(mouse_state->holding_object,dp);
							break;
						case 1: //drzim na prvi tocki
							constraint_manager->move_vertex_by(mouse_state->holding_object,1,dp);	
							break;
						case 2: //drzim na drugi tocki
							constraint_manager->move_vertex_by(mouse_state->holding_object,2,dp);	
							//mouse_state->holding_object->member->move_vertex_by(2,dp);
							break;
					}
					//zadevo narisem na novo	
					on_paint();
				}
				if (mouse_state->holding_constraint != NULL) {//pomeni da drzim v rokah nek objekt 
					if (mouse_state->holding_constraint->on_object(pozicija)) //ce je se na objektu
						mouse_state->holding_constraint->set_selected(1); //ga oznacim kot izbranega
					else
						mouse_state->holding_constraint->set_selected(0); //drugace kot neizbranega
					
					//zadevo narisem na novo
					on_paint();
				}
				break;
		}
	}
}

void Asaai::update_screen() {
	//zasezem handler od canvasa na katerega risem
	FXDCWindow dc(canvas);
	
	//nastimam barvo in debelino crte, ki jo risem na ekran
	dc.setForeground(FXRGB(mouse_state->color->r,mouse_state->color->g,mouse_state->color->b));
	dc.setLineWidth(mouse_state->width);

	//narsem v trace
	last_input->set_line(mouse_state->previous,mouse_state->current);
	//narisem se ne ekran
	dc.drawLine(mouse_state->previous->x,mouse_state->previous->y,mouse_state->current->x,mouse_state->current->y);
}

void Asaai::on_paint() {
	list <Object*>::iterator iter_obj;
	//zasezem handler od canvasa na katerega risem
	FXDCWindow dc(canvas);

	for (iter_obj = object_list.begin(); iter_obj != object_list.end(); iter_obj++) //dokler so objekti
	{
		if ((*iter_obj)->is_selected()) //ce je objekt izbran , potem mu uprajtam barvo in debelino
		{
			(*iter_obj)->set_object_color(mouse_state->color);
			(*iter_obj)->set_object_width(mouse_state->width);
		}
	}

	//pobrisem okno
	dc.setForeground(FXRGB(255,255,255));
	dc.fillRectangle(0,0,canvas->getWidth(),canvas->getHeight());
	
	if (show_constraints == 1) { //omejitve so vidne ... torej jih narisem
		constraint_manager->draw_constraints(&dc);
	}

	//risem objekte
	constraint_manager->draw_objects(&dc);
}


