/* recognizer.cpp
 * Copyright (C) 2004 Djukic Marko, Pernus Klemen, Zagar Miha
 * For conditions of distribution and use, see the copyright notice in main.cpp
 */
#include <iostream>
#include <list>

#define SQR(x) ((x)*(x))

#include "recognizer.h"
#include "math.h"

Recognizer::Recognizer(AI *_ai) {
	ai = _ai;
	recognized_objects = new LList<Object>();
}

unsigned char Recognizer::recognize(Trace *user_data) {

	get_points(user_data);

	Line *line = evaluate(ai->max_average_diff, user_data);
	if (line != NULL) {
		recognized_objects->add(line);
		//cekiraj za constrainte
	}
	else
		find_complex_lines();
	return 1;
}

Object *Recognizer::get_object() {
	LListMember<Object> *list_member;
	Object* objekt;

	list_member = recognized_objects->get_first(); //postavi na first
	if (list_member != NULL) { //obstaja objekt v listi
		objekt = list_member->member;
		list_member->del(0); //ga izbrisem iz seznama (objekt pustim pri miru)
	} else { //objekta ni
		objekt = NULL;
	}

	return objekt;

}

void Recognizer::get_points(Trace *user_data) {
	unsigned int num_of_points = user_data->get_nr_points();
	unsigned int i;
 
	points_list.clear();
	for(i = 0; i < num_of_points; i++) {
		points_list.push_back(new Point(user_data->get_point(i))); 
	}
}

Line* Recognizer::evaluate(float max_average_diff, Trace *user_data) {
	list <Point*>::iterator iter;

	int center_x = (int) (user_data->get_min_x() + (int)user_data->get_max_x() ) / 2;
	int center_y = (int) (user_data->get_min_y() + (int)user_data->get_max_y() ) / 2;

	float count = (float) user_data->get_nr_points();

	float sum_x = 0;
	float sum_y = 0;
	float sum_xy = 0;
	float sum_xx = 0;

	int x;
	int y;
	for(iter = points_list.begin(); iter != points_list.end(); iter++) {
		x = (*iter)->x - center_x;
		y = (*iter)->y - center_y;

		sum_x = sum_x + x; 
		sum_y = sum_y + y;
		sum_xy = sum_xy + x * y;
		sum_xx = sum_xx + SQR(x);
	}

	if ( count * sum_xx == SQR(sum_x) ) {
		cout << "Navpicna crta :)" << endl;
		Point start(user_data->get_min_x(), user_data->get_min_y());
		Point end(user_data->get_max_x(), user_data->get_max_y());
		return new Line(new Point(&start),new Point(&end));
	}

	float k = (count * sum_xy - sum_x * sum_y) / (count * sum_xx - SQR(sum_x));
	float c = (sum_y - k * sum_x) / count;
	
	float angle = (float) atan(k);
	cout << "y=" << k << "*x" << "+" << c << endl;

	float sum_diff = 0;

	for(iter = points_list.begin(); iter != points_list.end(); iter++) {
		x = (*iter)->x - center_x;
		y = (*iter)->y - center_y;

		sum_diff += dist_point_line(k, -1, c, new Point(x,y)); //a je tole memory leak?
	}

	float maxi = sum_diff/count;


	if (maxi < max_average_diff) { //tukile zamenjaj z verjetnostjo
		Point start(user_data->get_min_x(), user_data->get_min_y());
		Point end(user_data->get_max_x(), user_data->get_max_y());
		float len_half = (float) (dist_point_point(&start, &end) / 2);
		//Tukaj bi lahko uporabil novi tocki, vendar uporabim start in end
		start.x = (int) center_x + len_half * cos(angle);
		start.y = (int) center_y + len_half * sin(angle);
		end.x = (int) center_x - len_half * cos(angle);
		end.y = (int) center_y - len_half * sin(angle);
		return new Line(new Point(&start),new Point(&end));
	}
	else
		return NULL;
}



void Recognizer::find_complex_lines() {
	list <Point*> vl;
	list <Point*>::iterator iter_start; //pove odkod naj gledamo(za vse tocke)
	list <Point*>::iterator iter_end; //pove do kam naj gledamo (isto)
	list <Point*>::iterator vl_iter; //iterator za vl
	list <Point*>::iterator vl_iter_temp; //iterator za vl


	cout << "Find complex lines poklican!" << endl;
	vl.clear();
	vl = find_start_vertexes();

	vl_iter = vl.begin();
	iter_start = get_point_pos(vl_iter++, points_list.begin());
	iter_end = get_point_pos(vl_iter, iter_start);
	
	while (vl_iter != vl.end()) {
		if (perform_fitting(&vl, vl_iter, iter_start, iter_end, ai->complex_lines_1_epsilon, 0) == 1) {
			iter_start = iter_end;
			vl_iter++;
			if (vl_iter != vl.end())
				iter_end = get_point_pos(vl_iter, iter_start);
		}
		else {
			vl_iter--;
			iter_end = get_point_pos(vl_iter, iter_start);
		}

	}


	for(vl_iter = vl.begin(); vl_iter != vl.end(); vl_iter++) {
		iter_end = vl_iter;
		iter_end++;
		if (iter_end != vl.end()) 
			cout << "Crta:" << "("<<(*vl_iter)->x << "," << (*vl_iter)->y << ") " << "(" <<(*iter_end)->x << "," << (*iter_end)->y << ") "<< endl;
	}
	cout << "----------------------------------------" << endl;

	//Linearna regresija, pomaga uniciti odvecna oglisca
	vl_iter = vl.begin();
	if (vl.size() > 2) {
		iter_start = get_point_pos(vl_iter++, points_list.begin());
		iter_end = get_point_pos(++vl_iter, iter_start);
		vl_iter--;
		
		while (vl_iter != vl.end()) {
			vl_iter_temp = ++vl_iter;
			vl_iter--;
			if (perform_fitting(&vl, vl_iter, iter_start, iter_end, ai->complex_lines_2_epsilon, 1) == 0) {
				iter_start = get_point_pos(vl_iter++, iter_start);
				if (++vl_iter == vl.end())
					break;
				iter_end = get_point_pos(vl_iter, iter_start);
				vl_iter--;

			}
			else {
				vl_iter = vl_iter_temp;
				if (++vl_iter == vl.end())
					break;
				iter_end = get_point_pos(vl_iter, iter_start);
				vl_iter--;
			}
		}
	}
	//Cekiranje, kaksen kot tvorita 

	vl_iter = vl.begin();
	Point *prev, *curr, *next;
	prev = *vl_iter;
	vl_iter++;
	curr = *vl_iter;
	vl_iter++;
	float first, second;

	while (vl_iter != vl.end()) {
		if (vl_iter == vl.end())
			break;
		next = *vl_iter;

		
		first = atan2((float)prev->y - curr->y, (float)prev->x - curr->x);
		if (first < 0) first += 2* PI;
		second = atan2((float)next->y - curr->y, (float)next->x - curr->x);
		if (second < 0) second += 2* PI;

		cout << "Kot " << abs(first - second) * 180/3.14 - 180 << endl;
		
		if ( abs(abs(first - second) * 180 / PI - 180) < 20/*ai->get_angle_type_1_epsilon()*/ ) {
			curr = next;
			vl_iter_temp = vl_iter;
			vl_iter--;
			vl.erase(vl_iter);
			vl_iter = vl_iter_temp;
			cout << "Zbrisal sem ene stvar" << endl;

		}
		else {
			prev = curr;
			curr = next;

		}
		vl_iter++;
	}

	//Meatanje ven crt, ki so krajse kot 5 pikslov

	vl_iter = vl.begin();
//	vl_iter++;
	Line *line;

	while (1) {
		iter_start = vl_iter;
		iter_end = vl_iter;
		iter_end++;
		if (iter_end == vl.end())
			break;
		
		line = new Line(*iter_start, *iter_end);

		if ( line->get_length() < 5 ) {
			vl.erase(iter_end);
			cout << "Zbrisal sem eno stvar, ker je bila crta prekratka." << endl;
		}
		else
			vl_iter++;
	}

	for(vl_iter = vl.begin(); vl_iter != vl.end(); vl_iter++) {
		iter_end = vl_iter;
		iter_end++;
		if (iter_end != vl.end()) {
			recognized_objects->add(new Line(new Point(*vl_iter), new Point(*iter_end)));
		}

	}
	for(vl_iter = vl.begin(); vl_iter != vl.end(); vl_iter++) {
		iter_end = vl_iter;
		iter_end++;
		if (iter_end != vl.end()) 
			cout << "Crta:" << "("<<(*vl_iter)->x << "," << (*vl_iter)->y << ") " << "(" <<(*iter_end)->x << "," << (*iter_end)->y << ") "<< endl;
	}

}

list <Point*> Recognizer::find_start_vertexes() {
	list <Point*> vl;
	list <Point*>::iterator vl_iter;
	list <Point*>::iterator vl_iter_temp;

	list <Point*>::iterator iter_start;
	list <Point*>::iterator iter_end;
	list <Point*>::iterator iter;

	int x, y;

	int minX = 20000; //ena vrednost, ki nikoli ne bo 
	int minY = 20000;
	int maxX = -1;
	int maxY = -1;

	iter_start = points_list.begin();
	iter_end = points_list.end();
	for ( iter = iter_start; iter != iter_end; iter++ ) {
		x = (*iter)->x;
		y = (*iter)->y;
		
		if (x > maxX)
			maxX = x;
		if (y > maxY)
			maxY = y;
		if (x < minX)
			minX = x;
		if (y > maxY)
			maxY = y;
	}

	iter_start = points_list.begin();
	iter_end = points_list.end();
	iter_start++; //zacetno in koncno tocko dodamo posebej
	iter_end--;

	for ( iter = iter_start; iter != iter_end; iter++ ) {
		x = (*iter)->x;
		y = (*iter)->y;

		if (minY == y) 
			vl.push_back(new Point(x,y));
		else {
			if (maxY == y)
				vl.push_back(new Point(x,y));
			else { 
				if (minX == x)
					vl.push_back(new Point(x,y));
				else {
					if (maxX == x)
						vl.push_back(new Point(x,y));
				}
			}
		}
	}
	iter_start--;
	vl.push_front(new Point((*iter_start)->x,(*iter_start)->y)); //se zacetni tocki
	vl.push_back(new Point((*iter_end)->x,(*iter_end)->y));
	
	vl_iter = vl.begin();
	while (vl_iter != vl.end()) {  //mislim, da dela 100%
		vl_iter_temp = ++vl_iter; //tempu damo njegovo vrdnost +1
		vl_iter--; //njega nastimamo nazaj
		while ( ( vl_iter_temp != vl.end() ) && ( (*vl_iter_temp)->y == (*vl_iter)->y ) ) //gremo po istih y do konca ali do konca liste
			vl_iter_temp++;
		
		vl_iter++;

		if ( !(vl_iter == vl_iter_temp) ) { //ce smo se premaknili z zgornjo zanko vsaj za 1 (sicer je v redu)
			vl_iter_temp--;
			vl_iter = vl.erase(vl_iter, vl_iter_temp);
			vl_iter++;
		}
	} 

	//Isto kot zgoraj; vendar za x.
	vl_iter = vl.begin();
	while (vl_iter != vl.end()) {  //mislim, da dela 100%
		vl_iter_temp = ++vl_iter; //tempu damo njegovo vrdnost +1
		vl_iter--; //njega nastimamo nazaj
		while ( ( vl_iter_temp != vl.end() ) && ( (*vl_iter_temp)->x == (*vl_iter)->x ) ) //gremo po istih x do konca ali do konca liste
			vl_iter_temp++;
		
		vl_iter++;

		if ( !(vl_iter == vl_iter_temp) ) { //ce smo se premaknili z zgornjo zanko vsaj za 1 (sicer je v redu)
			vl_iter_temp--;
			vl_iter = vl.erase(vl_iter, vl_iter_temp);
			vl_iter++;
		}
	}

	return vl;
}

unsigned char Recognizer::perform_fitting(list <Point*> *vl, list <Point*>::iterator vl_iter,list <Point*>::iterator iter_start, list <Point*>::iterator iter_end, float max_aver_dist, unsigned char mode) {
	list <Point*>::iterator iter;
	Point* point; 
	Point* max_point = NULL;
	float a;
	float b;
	float c;
	int center_x;
	int center_y;
	int count = 0;
	
	//Mode 1 = linearna regrsija, Mode 0 = normalna metoda
	if (mode == 0) {
		//Premica (implicitna oblika zaradi x = const (ax+by+c=0)) skozi start in end
		a = (float) (*iter_end)->y - (*iter_start)->y;
		b = - (float) (*iter_end)->x + (*iter_start)->x;
		c = - (float) (*iter_start)->x * a - (*iter_start)->y * b;
	}
	else {
		center_x = 400;//(int) (user_data->get_min_x() + (int)user_data->get_max_x() ) / 2;
		center_y = 300;//(int) (user_data->get_min_y() + (int)user_data->get_max_y() ) / 2;

		float sum_x = 0;
		float sum_y = 0;
		float sum_xy = 0;
		float sum_xx = 0;

		int x;
		int y;
		for(iter = iter_start; iter != iter_end; iter++) {
			x = (*iter)->x - center_x;
			y = (*iter)->y - center_y;

			sum_x = sum_x + x; 
			sum_y = sum_y + y;
			sum_xy = sum_xy + x * y;
			sum_xx = sum_xx + SQR(x);
			count++;
		}
	
		if ( count * sum_xx == SQR(sum_x) ) {
			cout << "Navpicna crta :)" << endl;
		}

		float k = (count * sum_xy - sum_x * sum_y) / (count * sum_xx - SQR(sum_x));
		c = (sum_y - k * sum_x) / count;
	
		a = k;
		b = -1;
	}

	float mod = (float) sqrt(SQR(a)+SQR(b)); 
	//cout << a << " x + " << b << " y + " << c << endl;

	float max_dist = -1;
	float sum_dist = 0;
	float dist;
	count = 0;

	for (iter = iter_start; iter != iter_end; iter++) { //tukile morda porpavek iter_end
		point = *iter;

		if (mode == 0)
			dist = (1/mod) * abs(a * point->x + b * point->y + c); //razdalja tocke do premice
		else
			dist = (1/mod) * abs(a * (point->x - center_x) + b * (point->y - center_y) + c); //razdalja tocke do premice
		if (dist > max_dist) { 
			max_dist = dist;
			max_point = point;
		}

		sum_dist += dist; 
		count++;
	}
	if (mode == 1)
		cout << sum_dist/count << endl;
	
	if (sum_dist/count < max_aver_dist) {
		//Vse je v redu, naleteli smo na ravno crto
		if (mode == 1) {
			vl->erase(vl_iter);
		}
		return 1;
	}
	else {
		//Vrinemo novo oglisce na mesto, kamor kaze vl_iter	z vrednostjo max_point
		if (mode == 0) {
			if (max_point != NULL) {
				vl->insert(vl_iter, new Point(max_point->x, max_point->y));
			}
			else
			{
				cout << "PRISLO JE DO NAPAKE V GETPOINTPOS" << endl;
				return 1; //nismo vrinili nic
			}
		}
		return 0;
	}

}

list <Point*>::iterator Recognizer::get_point_pos(list <Point*>::iterator vl_iter, list <Point*>::iterator iter_start) {
	list <Point*>::iterator iter;

	iter = iter_start;
	while ( iter != points_list.end() ) {
		if (  ( (*iter)->x == (*vl_iter)->x ) && ( (*iter)->y == (*vl_iter)->y ) )
			return iter;
		iter++;
	}

}
float Recognizer::dist_point_line(float a, float b, float c, Point *point) {

	float mod = (float) sqrt(SQR(a)+SQR(b)); 
	return (float)(1/mod) * abs(a * point->x + b * point->y + c);
}

float Recognizer::dist_point_point(Point *point_1, Point *point_2) {

	return (float)sqrt( (float)SQR(point_1->x - point_2->x) + SQR(point_1->y - point_2->y) ) ;
}
