/* trace.cpp
 * Copyright (C) 2004 Djukic Marko, Pernus Klemen, Zagar Miha
 * For conditions of distribution and use, see the copyright notice in main.cpp
 */
#include "trace.h"
#include <malloc.h>
#include <windows.h>

#define ABS(a) (((a) < 0) ? -(a) : (a)) 



Trace::Trace() {
	trace_array_len = 0;

	trace_array = NULL;//(unsigned int *)malloc(sizeof(unsigned int) * trace_array_len); 

	nr_points = 0;
	min_x = max_x = min_y = max_y = 0;
}

Trace::~Trace() {
	free(trace_array);
}


void Trace::set_point(Point *tocka) {
	if (nr_points == 0) { //ce je to prva tocka ki jo dodajam podatvim meje
		max_x = min_x = tocka->x;
		max_y = min_y = tocka->y;
	} else { //popraim koordinate
		if (tocka->x > max_x)
			max_x = tocka->x;
		if (tocka->y > max_y)
			max_y = tocka->y;
		if (tocka->x < min_x)
			min_x = tocka->x;
		if (tocka->y < min_y)
			min_y = tocka->y;
	}
	
	nr_points++; //povecam stevilo tock
	//povecam prostor za novo tocko
	trace_array = (Point*)realloc(trace_array,sizeof(Point)*nr_points);
	//dodam tocko v array
	trace_array[nr_points-1].x = tocka->x;
	trace_array[nr_points-1].y = tocka->y;

}

void Trace::set_line(Point *t1, Point *t2) {
	//Algorithm credits to Po-Han Lin http://www.edepot.com
	bool yLonger=false;
	int incrementVal, endVal;
	int shortLen= t2->y - t1->y;
	int longLen= t2->x - t1->x;
	int decInc;
	int j = 0;
	
	if (ABS(shortLen) > ABS(longLen)) {
		int swap = shortLen;
		shortLen = longLen;
		longLen = swap;
		yLonger = true;
	}

	endVal = longLen;
	
	if (longLen < 0) {
		incrementVal = -1;
		longLen = -longLen;
	} else 
		incrementVal = 1;

	if (longLen == 0) 
		decInc = 0;
	else 
		decInc = (shortLen << 16) / longLen;

	if (yLonger) {	
		for (int i = 0; i != endVal; i +=  incrementVal) {
			set_point(new Point(t1->x + (j >> 16), t1->y + i));	
			j += decInc;
		}
	} else {
		for (int i = 0; i != endVal; i += incrementVal) {
			set_point(new Point(t1->x + i, t1->y + (j >> 16)));
			j += decInc;
		}
	}
}


Point *Trace::get_point(unsigned int index) {
	if (index < nr_points) { //ce index v mejah
		return trace_array + index;
	}

	return NULL;
}

Point *Trace::get_start() {
	if (nr_points > 0) 
        return trace_array;

	return NULL;
}

Point *Trace::get_end() {
	if (nr_points > 0)
		return trace_array + (nr_points-1);

	return NULL;
}

unsigned int Trace::get_min_x() {
	return min_x;
}

unsigned int Trace::get_min_y() {
	return min_y;
}

unsigned int Trace::get_max_x() {
	return max_x;
}

unsigned int Trace::get_max_y() {
	return max_y;
}

unsigned int Trace::get_nr_points() {
	return nr_points;
}


